#!/usr/bin/env python
#  _________________________________________________________________________
#
#  PyUtilib: A Python utility library.
#  Copyright (c) 2008 Sandia Corporation.
#  This software is distributed under the BSD License.
#  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
#  the U.S. Government retains certain rights in this software.
#  _________________________________________________________________________
#
from logging import getLogger, StreamHandler, INFO
from formatter import NullWriter
from subprocess import Popen, PIPE

class SvnError(Exception):
    """Exception raised due to svn error."""

class PyutilibSvnFormatter(NullWriter):
    maxLvlLength = 7 # <- compute this?
    field = "%%-%is %%s" % (maxLvlLength+2)
    fill = "\n" + ' '*(maxLvlLength+2+1)

    def format(self, record):
        msg = record.getMessage().replace("\n", PyutilibSvnFormatter.fill)
        level = "(" + record.levelname + ")"
        return PyutilibSvnFormatter.field % ( level, msg )

log = getLogger('pyutilib.svn')
log_handler = StreamHandler()
log_handler.setFormatter(PyutilibSvnFormatter())
log.addHandler(log_handler)
log.setLevel(INFO)

SvnVersion = (0,0)
try:
    p = Popen( [ 'svn', '--version', '--quiet' ],  stdout=PIPE, stderr=PIPE )
    stdout, stderr = p.communicate()
    if p.returncode:
        raise SvnError( "svn --version returned error %i:\n%s" %
                        (p.returncode, stderr) )
    SvnVersion = ( int(stdout.split('.')[0]),  int(stdout.split('.')[1]) )
except SvnError:
    raise
except:
    raise SvnError("Error determining svn version: is subversion on your path?")

def VerifySvnVersion(minVersion):
    for i in range(len(minVersion)):
        if len(SvnVersion) <= i:
            return False
        elif minVersion[i] < SvnVersion[i]:
            return True
        elif minVersion[i] > SvnVersion[i]:
            return False
    return True



#!/usr/bin/env python
#  _________________________________________________________________________
#
#  PyUtilib: A Python utility library.
#  Copyright (c) 2008 Sandia Corporation.
#  This software is distributed under the BSD License.
#  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
#  the U.S. Government retains certain rights in this software.
#  _________________________________________________________________________
#
from os import path
import re
import sys

from subprocess import Popen, PIPE
from pyutilib.svn.core import log, SvnError

dir_re = re.compile('\s+(\d+)\s+\S+\s+\w{3}\s+\d{2}\s+\S+\s+(\S+)')
externals_section_re = re.compile('^(\S+)\s+-(?:\s+(.*))?$')
proto = '(?:(?:svn|svn\+[sr]sh|file|https?)://)'
externals_14link_re = re.compile('^\s*((?!-r)\S+)\s+(?:-r\s*(\d+)\s+)?('+proto+'\S+)')
externals_15link_re = re.compile('^\s*(?:-r\s*(\d+)\s+)?((?:'+proto+'|(?:\.\.|\^)?/)\S+?)(?:\@(\d+))?\s+((?!'+proto+')\S+)')
log_revision_re = re.compile('^\s*r(\d+)')

# the list of subdirectories we will look for in a project
common_project_dirs = [ 'trunk', 'tags', 'branches', 
                        'stable', 'releases', 'maintenance', ]

# The subset of common_project_dirs that can be treated "normally" (as a
# collection of tags/branch subdirectories)
standard_subdirs = common_project_dirs[:]
standard_subdirs.remove('trunk')
standard_subdirs.remove('stable')


def svn_info(url, rev=None):
    info = {}
    try:
        is_int = int(rev) == rev
        if is_int:
            url += '@' + str(rev)
    except:
        pass
    cmd = ['svn', 'info', url]
    if rev:
        cmd.extend(['-r' + str(rev)])
    p = Popen(cmd, stdout=PIPE, stderr=PIPE)
    stdout, stderr = p.communicate()
    if p.returncode:
        raise SvnError( "svn info returned error %i:\n%s" %
                        (p.returncode, stderr) )
    for line in stdout.split('\n'):
        if line:
            key, val = line.split(':', 1)
            info[key.lower()] = val.strip()

    # special cases
    for x in ['revision', 'last changed rev']:
        if x in info:
            info[x] = int(info[x])
    return info


def svn_subdirs(url, subdir=None):
    _dir = {}
    if subdir:
        url = url.rstrip('/') + '/' + subdir.lstrip('/')

    p = Popen(['svn', 'ls', '-v', url], stdout=PIPE, stderr=PIPE)
    stdout, stderr = p.communicate()
    if p.returncode:
        raise SvnError( "svn ls returned error %i:\n%s" %
                        (p.returncode, stderr) )
    for line in stdout.split('\n'):
        g = dir_re.match(line)
        if not g:
            continue
        path = g.group(2).strip()
        if path[-1] != '/':
            raise SvnError("[internal] directory doesn't end with "
                           "/: our understanding of svn is wrong!")
        _dir[path[0:-1]] = int(g.group(1))
    return _dir


def svn_propget(property, url, rev=None):
    cmd = ['svn', 'propget']
    if type(property) in (list, tuple):
        cmd.extend(property)
    else:
        cmd.append(property)
    if rev:
        cmd.append(url + '@' + str(rev))
    else:
        cmd.append(url)
    log.debug(' '.join(cmd))
    p = Popen( cmd, stdout=PIPE, stderr=PIPE )
    stdout, stderr = p.communicate()
    if p.returncode:
        raise SvnError( "svn propget returned error %i:\n%s" %
                        (p.returncode, stderr) )
    return stdout.split('\n')


def svn_externals(url, rev=None):
    prop_data = svn_propget(['-R','svn:externals'], url, rev)
    svn15format = False
    externals = []
    srcdir = ''
    for line in prop_data:
        line = line.strip()
        g = externals_section_re.match(line)
        if g:
            assert g.group(1).startswith(url)
            srcdir = g.group(1)[len(url):].lstrip('/')
            if srcdir:
                srcdir = srcdir + '/'
            if g.group(2):
                line = g.group(2).strip()
            else:
                line = ""
        if line.startswith('#'):
            continue
        if len(line) == 0:
            continue
        g = externals_14link_re.match(line)
        if g:
            externals.append( External( srcdir+g.group(1), g.group(3),
                                        g.group(2) ) )
            continue
        g = externals_15link_re.match(line)
        if g:
            svn15format = True
            externals.append( External( srcdir+g.group(4), g.group(2),
                                        g.group(3) or g.group(1) ) )
            continue
        log.error("Unrecognized external format:\n     %s\n     found in %s" \
                  % (line, srcdir))

    return externals, svn15format


def svn_branch_created(url, rev=None):
    cmd = ['svn', 'log', '-q', '--stop-on-copy']
    if rev:
        cmd.append(url + '@' + str(rev))
    else:
        cmd.append(url)
    log.debug(' '.join(cmd))
    p = Popen( cmd, stdout=PIPE, stderr=PIPE )
    stdout, stderr = p.communicate()
    if p.returncode:
        raise SvnError( "svn log returned error %i:\n%s" %
                        (p.returncode, stderr) )
    for line in stdout.split('\n'):
        g = log_revision_re.match(line)
        if g:
            source_rev = g.group(1)
    return int(source_rev)


class External(object):
    def __init__(self, path, url, rev):
        self.path = path
        self.url = url
        self.revision = rev and int(rev) or None

    def __str__(self):
        return "%s: %s%s" % ( self.path, self.url, self.revision and
                              (" @ %s" % self.revision) or "" )


class Target(object):
    def __init__(self, project, name, rev):
        self.revision = 0
        self.create_rev = -1
        self.name = name
        self.url = project.url.rstrip('/') + '/' + name.lstrip('/')
        self.exernals = []
        self.svn15format = False
        self.update(rev)

    def __str__(self):
        ans = "%s @ %i : %s" % (self.name, self.revision, self.url)
        for e in self.externals:
            ans += "\n   %s" % str(e)
        return ans

    def update(self, new_rev):
        new_rev = int(new_rev)
        if self.revision == new_rev:
            return
        log.info("updating externals for target " + self.name)
        self.revision = new_rev
        self.create_rev = svn_branch_created(self.url)
        self.externals, self.svn15format = svn_externals(self.url)


class Project(object):
    def __init__(self, repo, path, _dir):
        self.revision = int( _dir['.'] )
        self.repo_url = repo.url
        self.url = repo.url.rstrip('/') + '/' + path.lstrip('/')
        # the project name is either the repo name or the subdirectory name...
        self.name = self.url.rstrip('/').split('/')[-1]
        self.targets = {}
        self.update(_dir)

    def update(self, _dir):
        old_targets = self.targets;
        self.targets = {}
        subdirs = _dir.keys()
        self.revision = int( _dir['.'] )

        log.info("updating project %s" % self.name)

        # Handle trunk... this is easy
        if 'trunk' in subdirs:
            self._record_target(old_targets, 'trunk', _dir['trunk']);

        # "stable" is tricky: it can either be a copy of trunk, OR it
        # can be a collection of stable branches, we will assume that if
        # ALL subdirectories within stable either start or end with a
        # number, then the project supports multiple stable tags
        if 'stable' in subdirs:
            stable_dirs = svn_subdirs(self.url, 'stable')
            if len(stable_dirs) == 1: # 1 is the '.' directory!
                # default to multiple stable dirs if stable is empty
                multiple_stable = True
            else:
                ver_count = 0.0
                for d in stable_dirs.keys():
                    if d == '.':
                        continue
                    if d.strip('1234567890.') != d:
                        ver_count += 1

                common_count = 0.0
                if 'trunk' in subdirs:
                    trunk_dirs = svn_subdirs(self.url, 'trunk').keys()
                    for d in stable_dirs.keys():
                        if d == '.':
                            continue
                        if d in trunk_dirs:
                            common_count += 1

                total = float( len(stable_dirs) - 1 )
                if ( ver_count/total > 0.25 ) ^ ( common_count/total < 0.25 ):
                    log.warn("I am confused as to if stable is a branch "\
                             "or collection of branches; assuming a "\
                             "collection of branches")
                    multiple_stable = True
                else:
                    multiple_stable = bool( ver_count/len(stable_dirs) > 0.25 )
            if multiple_stable:
                for key, rev in stable_dirs.items():
                    if key == '.':
                        continue
                    self._record_target(old_targets, 'stable/'+key, rev)
            else:
                self._record_target(old_targets, 'stable', _dir['stable'])

        # handle the other common "collections" of branch- or tag-like
        # constructs
        for d in standard_subdirs:
            if d in subdirs:
                for key, rev in svn_subdirs(self.url, d).items():
                    if key == '.':
                        continue
                    self._record_target(old_targets, d+'/'+key, rev)

        # Assume everything else is a simple target.
        for d, rev in _dir.iteritems():
            if d in common_project_dirs or d == '.':
                continue
            self._record_target(old_targets, d, rev)


    def _record_target(self, old, name, rev):
        if name in old:
            self.targets.setdefault(name, old[name]).update(rev)
        else:
            self.targets[name] = Target(self, name, rev)

    def __str__(self):
        ans = "Project: %s\nRevision: %i\nURL: %s\nTargets:\n" % \
              (self.name, self.revision, self.url)

        target_list = self.targets.keys()
        target_list.sort()
        for t in target_list:
            ans = ans + ( "   %s\n" %
                          ( str(self.targets[t]).replace('\n', '\n   ') ) )
        return ans


class Repository(object):
    def __init__(self, url, rev=None):
        info = svn_info(url, rev)
        if 'repository root' not in info:
            raise SvnError("ERROR: Specified repository url (%s) does "
                           "not appear to exist" % url)
        self.url = info['repository root']
        self.revision = 0
        self.projects = {}

    def update(self):
        log.info("updating repository %s" % self.url)
        info = svn_info(self.url)
        if self.url != info['repository root']:
            raise SvnError("ERROR: Specified repository url (%s) does "
                           "not match repository root (%s)" %
                           (self.url, info['Repository Root']))

        # duck out if nothing has changed
        if self.revision == info['revision']:
            return False
        self.revision = info['revision']

        old_projects = self.projects
        self.projects = {}
        potential_projects = ['']
        while potential_projects:
            target = potential_projects.pop()
            _dir = svn_subdirs(self.url, target)
            subdirs = _dir.keys()
            project_found = False
            for d in common_project_dirs:
                if d in subdirs:
                    project_found = True
                    self._record_project(old_projects, target, _dir)
                    break
            if not project_found:
                subdirs.remove('.')
                potential_projects.extend([target+'/'+x for x in subdirs])

        # catch-all for repositories that don't honor trunk/tags/etc
        if len(self.projects) == 0:
            self._record_project(old_projects, '', svn_subdirs(self.url,''))

        return True

    def _record_project(self, old, name, _dir):
        name = name.strip('/')
        if name in old:
            self.projects.setdefault(name, old[name]).update(_dir)
        else:
            self.projects[name] = Project(self, name, _dir)


    def __str__(self):
        ans = "Repository: %s\nRevision: %i\nProjects:\n" % \
              (self.url, self.revision)

        project_list = self.projects.keys()
        project_list.sort()
        for p in project_list:
            ans = ans + ( "   %s\n" %
                          ( str(self.projects[p]).replace('\n', '\n   ') ) )
        return ans


if __name__ == "__main__":
    # Check for sane usage.
    if len(sys.argv) != 2:
        sys.stderr.write("Usage: %s REPOS_PATH\n"
                         % (path.basename(sys.argv[0])))
        sys.exit(1)

    r = Repository(sys.argv[1])
    r.update()
    print(r)



#!/usr/bin/env python
#  _________________________________________________________________________
#
#  PyUtilib: A Python utility library.
#  Copyright (c) 2008 Sandia Corporation.
#  This software is distributed under the BSD License.
#  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
#  the U.S. Government retains certain rights in this software.
#  _________________________________________________________________________
#
import os
import re
import sys

from datetime import datetime, timedelta
from subprocess import Popen, PIPE

from pyutilib.svn.repository import svn_info, svn_externals, SvnError


def checkout(path_list, date_str):
    target, url, rev = path_list.pop(0)

    # If the revision wasn't specified (i.e. fixed by the external
    # itself), then we need to determine the correct revision number
    # from the main date_str.
    if rev is None:
        # find the repo (the deep link may no longer exist) and get the rev
        tmp_url = url;
        while len(tmp_url):
            try:
                rev = svn_info(tmp_url, '{'+date_str+'}')['revision']
            except SvnError:
                pass
            if rev is not None:
                break
            try:
                tmp_url,junk = tmp_url.rsplit('/',1)
            except ValueError:
                tmp_url = ""
        if not tmp_url:
            raise SvnError( ("unable to determine repository revision number "
                             "for '%s' on %s" % (url, date_str) ) )

    # Check out the url to the target path, but DO NOT automatically
    # checkout the externals
    print("(INFO) Checking out %s at r%s to %s" % (url, rev, target))
    cmd = [ 'svn', 'co', '-r', str(rev), '--ignore-externals',
            url+'@'+str(rev), target ]
    #print(' '.join(cmd))
    p = Popen(cmd, stdout=PIPE, stderr=PIPE)
    stdout, stderr = p.communicate()
    if p.returncode:
        raise SvnError( "svn info returned error %i:\n%s" %
                        (p.returncode, stderr) )

    # Parse the repo and get any externals that still need to be checked
    # out.  Because we never change the working directory, we need to
    # prepend the current target path to the external path before pushing
    # the external onto the list of pending paths.
    print("(INFO) Parsing externals for %s" % (url))
    externals, svn15format = svn_externals(url, rev)
    while externals:
        tmp = externals.pop(0)
        path_list.append((target+'/'+tmp.path, tmp.url, tmp.revision))


#-----------------------------------------------------------------------
# MAIN SCRIPT
#

def main():
    # Check for sane usage.
    me = os.path.basename(sys.argv[0])
    if len(sys.argv) < 3 or len(sys.argv) > 4:
        print("Usage: %s {revision|date} repo_url [path]" % (me))
        print("")
        print("  revision : the revision number to warp back to")
        print("  date     : the date to warp back to (YYYY-MM-DD hh:mm:ss -TZ)")
        print("  repo_url : the repository url to check out")
        print("  path     : the destination path for the working copy")
        print("")
        sys.exit(1)

    rev_or_date = sys.argv[1]
    url = sys.argv[2]
    if len(sys.argv) > 3:
        target = sys.argv[3]
    else:
        target = url.rsplit('/',1)[1]

    # First, get the date that we will be using to interrogate the repository
    if re.match('^\d+$', rev_or_date):
        fields = svn_info(url, rev_or_date)['last changed date'] \
                 .split('(', 1)[0].strip().split(' ')

        # Because this is the time that the revision was actually committed,
        # we need to add a second to get the state immediately *after* the
        # commit.
        d = [int(x) for x in fields.pop(0).split('-')]
        f = []
        if len(fields):
            f = [int(x) for x in fields.pop(0).split(':')]
        while len(f) < 3:
            f.append(0)
        Date = datetime(d[0],d[1],d[2],f[0],f[1],f[2]) + timedelta(seconds=1)
        fields.insert(0, Date.isoformat(' '))
        date_str = ' '.join(fields)
    else:
        date_str = rev_or_date

    print("(INFO) checking out the repository as it existed on %s" % date_str)

    # Do the actual checkout -- this is recursive, so it will pick up ALL
    # externals.
    path_list = [(target, url, None)]
    while path_list:
        checkout(path_list, date_str)

if __name__ == '__main__':
    main()



#!/usr/bin/env python
#  _________________________________________________________________________
#
#  PyUtilib: A Python utility library.
#  Copyright (c) 2008 Sandia Corporation.
#  This software is distributed under the BSD License.
#  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
#  the U.S. Government retains certain rights in this software.
#  _________________________________________________________________________
#

import logging
import os
import re
import sys
import yaml
from optparse import OptionParser, IndentedHelpFormatter
from pyutilib.svn.core import log
from pyutilib.svn.database import RepositoryDatabase, DatabaseError
from pyutilib.svn.external_manager import ExternalManager

class CommandParserCommand(Exception):
    def __init__(self, msg):
        self.msg = msg
    def __str__(self):
        return self.msg

class CommandParser(OptionParser):
    def error(self, what):
        g = re.match('no such option: (.+)', what)
        if g:
            raise CommandParserCommand(g.group(1))
        OptionParser.error(self, what)

class DescriptionFormatter(IndentedHelpFormatter):
    def format_description(self, description):
        parts = description.split("\n\n")
        ans = ""
        for part in parts:
            tmp = IndentedHelpFormatter.format_description(self, part)
            if tmp and tmp[0] != " " and len(ans):
                ans += "\n" + tmp
            else:
                ans += tmp
        return ans


class ScriptCommand(object):
    def help(self, pm):
        pass
    def register_options(self, parser):
        pass
    def process(self, pm, options, args):
        pass


class help_cmd(ScriptCommand):
    def help(self, pm):
        return """ \
%prog: a utility for managing externals-based projects within
Subversion.  Specific help for individual subcommands is available by
running '%prog help <subcommand>'.

Available subcommands:\n\n   """ + "\n\n   ".join(sorted(pm.commands.keys()))

    def process(self, pm, options, args):
        if len(args) > 1:
            cmd = args[1]
            if cmd not in pm.commands:
                print("Unknown command: " + cmd)
                return 2
            info = cmd.upper() + ": " + ( pm.commands[cmd].help(pm) or
                                          "(no description provided)")
            parser = OptionParser(usage=pm.usage,
                                  description=info,
                                  formatter=DescriptionFormatter())
            pm.common_args(parser)
            pm.commands[cmd].register_options(parser)
            parser.print_help()
        else:
            self.process(pm, options, ['help', 'help'])


class CommandManager(object):
    def __init__(self):
        self.commands = {}
        self.register_command("help", help_cmd())
        self.usage = """usage: %prog subcommand [options] [args]
Type '%prog help <subcommand>' for help on a specific subcommand."""

    def register_command(self, name, cmd):
        if name in self.commands:
            raise ProjectManager_Error("duplicate command name: %s" % name)
        self.commands[name] = cmd

    def common_args(self, parser):
        parser.add_option('--verbose', '-v', action="count", dest='verbose',
                          default=CommandManager.default_verbosity(),
                          help="increase verbosity; "
                          "can be specified multiple times")
        parser.add_option('--quiet', '-q', action="store_true", dest='quiet',
                          help="silence all messages except for errors")

    @staticmethod
    def default_verbosity():
        return 1

    def process(self, args=None):
        if args is None:
            args = sys.argv[1:]
        else:
            args = args[:]

        # We need to preemptively parse the command line to try and
        # determine the command name
        parser = CommandParser(usage=self.usage)
        self.common_args(parser)
        try:
            options, largs = parser.parse_args(args[:])
        except CommandParserCommand:
            e = sys.exc_info()[1]
            if len(parser.largs):
                largs = parser.largs
            else:
                largs = [e.msg]

        if len(largs) == 0:
            log.error("No subcommand specified")
            parser.print_usage()
            return 2

        command = largs[0]
        if command not in self.commands:
            log.error("Unknown subcommand: " + command)
            parser.print_usage()
            return 2

        # now that we know the command, pars the args for real
        parser = OptionParser(usage=self.usage)
        self.common_args(parser)
        self.commands[command].register_options(parser)
        options, largs = parser.parse_args(args[:])

        if not len(largs) or largs[0] != command:
            log.error( "My understanding of the subcommand changed between "
                       "preprocessing (%s) and runtime (%s)" %
                       ( command, len(largs) and "None" or largs[0] ) )
            return 1

        # process common options
        if options.quiet or options.verbose == 0:
            log.setLevel(logging.ERROR)
        elif options.verbose == 1:
            log.setLevel(logging.WARNING)
        elif options.verbose == 2:
            log.setLevel(logging.INFO)
        elif options.verbose >= 3:
            log.setLevel(logging.DEBUG)

        # run the command
        return self.commands[command].process(self, options, largs)



class db_cmd(ScriptCommand):
    def help(self, pm):
        return """
Perform administration on the cached database of Subversion
repository information."""

    def register_options(self, parser):
        parser.add_option('--add', '-a', action="append", dest='add',
                          default=[],
                          help="add a new repository to the database")
        parser.add_option('--remove', action="append", dest='remove',
                          default=[],
                          help="remove a repository from the database")
        parser.add_option('--update', '-u', action="store_true", dest="update",
                          help="update the repository database cache")
        parser.add_option('--rebuild', action="store_true", dest="rebuild",
                          help="rebuild the repository database cache")
        parser.add_option('--print', '-p', action="store_true", dest="printdb",
                          help="print the repository database")
        parser.add_option('--list', '-l', action="append", dest="list",
                          default=[],
                          help="list the specified database components "\
                              "[repos, projects, targets]")

    def process(self, pm, options, args):
        db = RepositoryDatabase(options.db)
        try:
            db.load()
        except DatabaseError:
            log.info("cache file DNE; initializing empty database")

        for url in options.add:
            db.add(url)
            db.save()
        for url in options.remove:
            db.remove(url)
            db.save()
        if options.rebuild:
            db.rescan()
            db.save()
        if options.update:
            if db.update():
                db.save()
        if options.printdb:
            print(db)
        for component in options.list:
            if component == 'repos':
                print("\n".join(sorted(db.repos.keys())))
            elif component == 'projects':
                for r in sorted(db.repos.keys()):
                    print("\n".join(sorted([ x.url for x in
                                             db.repos[r].projects.values() ])))
            elif component == 'targets':
                for r in sorted(db.repos.keys()):
                    for p in sorted(db.repos[r].projects.keys()):
                        print("\n".join(sorted([x.url for x in
                                   db.repos[r].projects[p].targets.values()])))
            else:
                log.error("Unknown component " + component)


class update_cmd(ScriptCommand):
    def help(self, pm):
        return """
Recursively update all repositories in the repository database cache.
This subcommand will update all repositories in the database and then
recursively parse all externals in each repository and add the
repositories that those externals point to to the repository database.
Pay special attention to the 'aliases; and 'exclude' sections of the
configuration file to prevent repositories from being included multiple
times through different URLs, or to exclude repositories completely."""

    def register_options(self, parser):
        parser.add_option('--config', '-c', action="store", dest='config',
                          help="the location of the configuration file "
                          "(default=%s)" % ExternalManager.default_config())

    def process(self, pm, options, args):
        em = ExternalManager(options.config)
        em.update()


class check_cmd(ScriptCommand):
    def help(self, pm):
        return """
Check the integrity of all externals in the repository database.  This
will identify issues with externals, including:

  - externals that point through an aliased URL.

  - externals that point to an excluded URL.

  - externals that point to a location other than a known project target.

  - pegged externals that work but point to locations that no longer
exist at the HEAD.

  - externals that are pegged to revisions older than HEAD for that location.

  - externals that should be pegged, but are not.
"""

    def register_options(self, parser):
        parser.add_option('--config', '-c', action="store", dest='config',
                          help="the location of the configuration file "
                          "(default=%s)" % ExternalManager.default_config())
        parser.add_option('--peg', action="store_true", dest='peg',
                          help="validate that all externals referenced "
                          "through pegged targets are also pegged")
        parser.add_option('--key', '-k', action="store_true", dest='key',
                          help="print a key to all the symbols used")
        parser.add_option('--project', '-p', action="store", dest='project',
                          help="only check specified project")
        parser.add_option('--repo', '-r', action="store", dest='repo',
                          help="only check specified repository")

    def process(self, pm, options, args):
        em = ExternalManager(options.config)
        em.resolve_targets()
        proj = None
        repo = None
        if options.project:
            proj = em.get_project(options.project)
            if proj is None:
                log.error( "Specified project (%s) does not exist" %
                           options.project )
                return 1
        if options.repo:
            repo = em.get_repository(options.repo)
            if repo is None:
                log.error( "Specified repository (%s) does not exist" %
                           options.repo )
                return 1
        if options.peg:
            em.check_pegged(project=proj, repo=repo, key=options.key)
        else:
            em.check_integrity(project=proj, repo=repo, key=options.key)


class status_cmd(ScriptCommand):
    def help(self, pm):
        return """
Print information on the status for projects in the repository database.
"""

    def register_options(self, parser):
        parser.add_option('--config', '-c', action="store", dest='config',
                          help="the location of the configuration file "
                          "(default=%s)" % ExternalManager.default_config())
        parser.add_option('--project', '-p', action="store", dest='project',
                          help="only check specified project")
        parser.add_option('--repo', '-r', action="store", dest='repo',
                          help="only check specified repository")

    def process(self, pm, options, args):
        em = ExternalManager(options.config)
        em.resolve_targets()
        proj = None
        repo = None
        if options.project:
            proj = em.get_project(options.project)
            if proj is None:
                log.error( "Specified project (%s) does not exist" %
                           options.project )
                return 1
        if options.repo:
            repo = em.get_repository(options.repo)
            if repo is None:
                log.error( "Specified repository (%s) does not exist" %
                           options.repo )
                return 1
        em.print_status(repo=repo, project=proj)


class diff_cmd(ScriptCommand):
    def help(self, pm):
        return """
Print a difference between the specified repository definition and the
current state of the repository (as stored in the repository database).

usage: %prog diff repo_defn [options] [args]

   repo_defn : the YAML repository definition file to compare against
"""

    def register_options(self, parser):
        parser.add_option('--config', '-c', action="store", dest='config',
                          help="the location of the configuration file "
                          "(default=%s)" % ExternalManager.default_config())

    def process(self, pm, options, args):
        fname = len(args) > 1 and args[1] or None
        if not fname or not os.path.exists( fname ):
            log.error("Definition file (%s) does not exist" % str(fname))
            return 1

        em = ExternalManager(options.config)
        em.resolve_targets()
        user_description = em.load_repo_definition(fname)
        description = em.finalize_repo_definition(user_description)
        changes = em.generate_changeset_from_definition(description)
        em.print_changeset(changes)


class inspect_cmd(ScriptCommand):
    def help(self, pm):
        return """
Generate the YAML repository definition for the specified Subversion
repository.  It is HIGHLY recommended that you
update your repository database cache BEFORE running this command.

usage: %prog inspect repo_defn [options]

   repo_defn : the YAML repository definition file to generate
"""

    def register_options(self, parser):
        parser.add_option('--config', '-c', action="store", dest='config',
                          help="the location of the configuration file "
                          "(default=%s)" % ExternalManager.default_config())
        parser.add_option('--project', '-p', action="store", dest='project',
                          help="generate description for the specified project")
        parser.add_option('--repo', '-r', action="store", dest='repo',
                          help="generate description for the specified repository")

    def process(self, pm, options, args):
        fname = len(args) > 1 and args[1] or None
        if not fname:
            log.error("Definition file was not specified.")
            return 1
        if os.path.exists( fname ):
            log.error("Definition file (%s) exists" % str(fname))
            return 1

        em = ExternalManager(options.config)
        em.resolve_targets()

        proj = None
        repo = None
        if options.project:
            proj = em.get_project(options.project)
            if proj is None:
                log.error( "Specified project (%s) does not exist" %
                           options.project )
                return 1
        if options.repo:
            repo = em.get_repository(options.repo)
            if repo is None:
                log.error( "Specified repository (%s) does not exist" %
                           options.repo )
                return 1

        descrip = em.generate_repo_definition(proj=proj, repo=repo)
        #final = em.finalize_repo_definition(descrip)
        ans = em.collapse_repo_definition(descrip)
        log.info("Writing description to " + fname)
        FILE = open(fname, 'w')
        FILE.write( yaml.dump( ans, default_flow_style=False,
                               explicit_start=True, explicit_end=True,
                               width=160 ).replace('{}','') )
        FILE.close()


class apply_cmd(ScriptCommand):
    def help(self, pm):
        return """
Create a working copy and apply any changes necessary to match the
specified repository definition.  It is HIGHLY recommended that you
update your repository database cache BEFORE running this command.

usage: %prog apply repo_defn [options] [args]

   repo_defn : the YAML repository definition file to compare against
"""

    def register_options(self, parser):
        parser.add_option('--config', '-c', action="store", dest='config',
                          help="the location of the configuration file "
                          "(default=%s)" % ExternalManager.default_config())

    def process(self, pm, options, args):
        fname = len(args) > 1 and args[1] or None
        if not fname or not os.path.exists( fname ):
            log.error("Definition file (%s) does not exist" % str(fname))
            return 1

        em = ExternalManager(options.config)
        em.resolve_targets()
        user_description = em.load_repo_definition(fname)
        description = em.finalize_repo_definition(user_description)
        changes = em.generate_changeset_from_definition(description)
        em.implement_changeset(changes, description)



class branch_cmd(ScriptCommand):
    def help(self, pm):
        return """
Prepare a branch of an existing project.

usage: %prog branch repo_defn [options] [args]

   repo_defn : the YAML repository definition file to update with the
               branch configuration
"""

    def register_options(self, parser):
        parser.add_option('--config', '-c', action="store", dest='config',
                          help="the location of the configuration file "
                          "(default=%s)" % ExternalManager.default_config())
        parser.add_option('--project', '-p', action="store", dest='project',
                          help="project to branch")
        parser.add_option('--source', action="store", dest='source',
                          help="source target (target to be copied)")
        parser.add_option('--dest', action="store", dest='dest',
                          help="destination target (target to be created)")
        parser.add_option('--peg', action="store_true", dest='peg',
                          help="peg all externals to fixed revision numbers")
        parser.add_option('--dry-run', action="store_true", dest='dryrun',
                          help="configure branch, but do not add it to "
                          "the definition file")

    def process(self, pm, options, args):
        fname = len(args) > 1 and args[1] or None
        if not fname or not os.path.exists( fname ):
            log.error("Definition file (%s) does not exist" % str(fname))
            return 1

        em = ExternalManager(options.config)
        em.resolve_targets()

        branch_cfg = {'peg' : options.peg}
        if options.project:
            branch_cfg['project'] = options.project
        if options.source:
            branch_cfg['src'] = options.source
        if options.dest:
            branch_cfg['target'] = options.dest
        defn, proj_name = em.branch_config(branch_cfg)

        tmp = em.collapse_repo_definition(defn)
        log.info("Configured branch: \n" + yaml.dump(
            tmp, default_flow_style=False,
            explicit_start=True, explicit_end=True,
            width=160 ).replace('{}',''))

        if options.dryrun:
            return 0

        user_description = em.load_repo_definition(fname)
        if defn['__repo__'] != user_description['__repo__']:
            log.error("Definition file does not match branch repository")
            return 1
        user_description[proj_name].update(defn[proj_name])
        ans = em.collapse_repo_definition(user_description)
        log.info("Writing description to " + fname)
        FILE = open(fname, 'w')
        FILE.write( yaml.dump( ans, default_flow_style=False,
                               explicit_start=True, explicit_end=True,
                               width=160 ).replace('{}','') )
        FILE.close()

        #description = em.finalize_repo_definition(user_description)
        #changes = em.generate_changeset_from_definition(description)
        #em.print_changeset(changes)


class SVNPM_CommandManager(CommandManager):
    def __init__(self):
        CommandManager.__init__(self)
        self.register_command("db",      db_cmd())
        self.register_command("update",  update_cmd())
        self.register_command("check",   check_cmd())
        self.register_command("diff",    diff_cmd())
        self.register_command("inspect", inspect_cmd())
        self.register_command("apply",   apply_cmd())
        self.register_command("status",  status_cmd())
        self.register_command("branch",  branch_cmd())

    def common_args(self, parser):
        CommandManager.common_args(self, parser)
        parser.add_option('--database','-d', action="store", dest='db',
                          default=RepositoryDatabase.default_db(),
                          help="the repository database file "
                          "(default=%default)")


def main():
    pm = SVNPM_CommandManager()
    sys.exit(pm.process())

if __name__ == '__main__':
    main()
